<?php
/* --------------------------------------------------------------
   GMPayPal.php 2.4_2014-08-19 gambio
   Gambio GmbH
   http://www.gambio.de
   Copyright (c) 2014 Gambio GmbH
   Released under the GNU General Public License (Version 2)
   [http://www.gnu.org/licenses/gpl-2.0.html]
   --------------------------------------------------------------
*/

class GMPayPal {
	protected $_config = array(
		'username' => '',
		'password' => '',
		'signature' => '',
		'username_sandbox' => '',
		'password_sandbox' => '',
		'signature_sandbox' => '',
		'sandbox_mode' => 1,
		'orders_status_id_tmp' => 0,
		'orders_status_id_pending' => 0,
		'orders_status_id_authorized' => 1,
		'orders_status_id_ordered' => 1,
		'orders_status_id' => 1,
		'paymentaction' => 'Sale',
		'use_paypal_checkout' => 1,
		'use_paypal_products' => 1,
		'order_max_auth' => 10,
		'order_max_capture' => 10,
		'button_style' => 'Sunrise', # Sunrise | Silver
		'prevent_selfpickup' => 'prevent',
		'cppheaderimage' => '',
		'cpplogoimage' => '',
		'cppheaderbordercolor' => '000000',
		'cppheaderbackcolor' => 'ffffff',
		'cpppayflowcolor' => 'ff0000',
		'cppcartbordercolor' => '000000',
		'pagestyle' => '',
		'brandname' => STORE_NAME,
	);
	protected $_logger;
	protected $_debug = true;
	protected $_txt;

	const DB_KEY_PREFIX = 'PAYPAL_';
	const LOGLEVEL_INFO = 0;
	const LOGLEVEL_WARN = 10;
	const LOGLEVEL_ERROR = 20;
 	const LOGLEVEL_DEBUG = 99;
 	const MODULE_VERSION = '2.4_2014-04-22';

	// StateOrProvince req'd for Argentina, Brazil, Canada, China, Indonesia, India, Japan, Mexico, Thailand, USA
	// state data missing: AR, CN, ID, TH
 	const STATEORPROVINCEREQUIRED = 'AR,BR,CA,CN,ID,IN,JP,MX,TH,US';
	
	public function __construct() {
		if(class_exists('LogControl')) {
			$this->_logger = LogControl::get_instance();
		}
		else {
			$this->_logger = false;
		}
		$this->_txt = new LanguageTextManager('paypalng', $_SESSION['languages_id']);
		$config = $this->getConfigArray();
		if(!defined('PP_CONFIG_PATH')) {
			if($config['sandbox_mode'] == 1) {
				define('PP_CONFIG_PATH', DIR_FS_CATALOG.'ext/paypal-sdk/config');
			}
			else {
				define('PP_CONFIG_PATH', DIR_FS_CATALOG.'ext/paypal-sdk/config-live');
			}
		}
		//$this->_log('pp_config in '. PP_CONFIG_PATH);
		require_once DIR_FS_CATALOG.'ext/paypal-sdk/PPAutoloader.php';
		PPAutoloader::register();
	}
	
	public function _log($message, $loglevel = self::LOGLEVEL_INFO) {
		if($this->_logger !== false) {
			switch($loglevel)
			{
				case self::LOGLEVEL_ERROR:
					$this->_logger->error($message, 'payment', 'payment.paypal');
					break;
				case self::LOGLEVEL_WARN:
					$this->_logger->warning($message, 'payment', 'payment.paypal');
					break;
				case self::LOGLEVEL_INFO:
				case self::LOGLEVEL_DEBUG:
				default:
					$this->_logger->notice($message, 'payment', 'payment.paypal');
					break;
			}
		}
	}

	public function get_text($placeholder) {
		return $this->_txt->get_text($placeholder);
	}

	public function replaceLanguagePlaceholders($content) {
		while(preg_match('/##(\w+)\b/', $content, $matches) == 1) {
			$replacement = $this->get_text($matches[1]);
			if(empty($replacement)) {
				$replacement = $matches[1];
			}
			$content = preg_replace('/##'.$matches[1].'/', $replacement.'$1', $content, 1);
		}
		return $content;
	}

	public static function _formatAmount($amt) {
		$formatted = number_format($amt, 2, '.', ',');
		return $formatted;
	}

	public function getConfigArray() {
		foreach($this->_config as $key => $value) {
			$db_key = self::DB_KEY_PREFIX.strtoupper($key);
			$dbvalue = gm_get_conf($db_key);
			if($dbvalue != null) {
				$this->_config[$key] = $dbvalue;
			}
		}
		return $this->_config;
	}

	public function setConfigValue($key, $value) {
		$db_key = self::DB_KEY_PREFIX.strtoupper($key);
		$this->_log("setting $db_key to $value");
		$value = xtc_db_input($value);
		gm_set_conf($db_key, $value);
	}

	public function isPayPalActive() {
		$module_installed = strpos(MODULE_PAYMENT_INSTALLED, 'paypalng.php') !== false;
		$module_active = defined('MODULE_PAYMENT_PAYPALNG_STATUS') && strtolower(MODULE_PAYMENT_PAYPALNG_STATUS) == 'true';
		$is_active = $module_installed && $module_active;
		return $is_active;
	}

	public function isExpressCheckoutActive() {
		$config = $this->getConfigArray();
		$active = $config['use_paypal_checkout'] == true && $this->isPayPalActive();
		return $active;
	}

	public function isProductsPageButtonActive() {
		$config = $this->getConfigArray();
		$active = $config['use_paypal_products'] == true && $this->isPayPalActive();
		return $active;
	}

	protected function _getCredentials() {
		$config = $this->getConfigArray();
		if($config['sandbox_mode'] == 1) {
			$cred = new PPSignatureCredential($config['username_sandbox'], $config['password_sandbox'], $config['signature_sandbox']);
		}
		else {
			$cred = new PPSignatureCredential($config['username'], $config['password'], $config['signature']);
		}
		return $cred;
	}

	public static function orderHasTransactions($orders_id) {
		$query = "SELECT COUNT(*) AS num FROM paypal_transactions WHERE orders_id = ".(int)$orders_id;
		$result = xtc_db_query($query);
		$has_transactions = false;
		while($row = xtc_db_fetch_array($result)) {
			$has_transactions = $row['num'] > 0;
		}
		return $has_transactions;
	}

	public function getPaycode($orders_id) {
		$query = "SELECT * FROM paypal_paylink WHERE orders_id = ".(int)$orders_id;
		$result = xtc_db_query($query);
		$paycode = false;
		while($row = xtc_db_fetch_array($result)) {
			$paycode = $row;
		}
		return $paycode;
	}

	public function getPaycodeByCode($paycode) {
		$query = "SELECT * FROM paypal_paylink WHERE paycode = '".xtc_db_input($paycode)."'";
		$result = xtc_db_query($query);
		$paycode = false;
		while($row = xtc_db_fetch_array($result)) {
			$paycode = $row;
		}
		return $paycode;
	}

	public function deletePaycode($orders_id) {
		xtc_db_query("DELETE FROM paypal_paylink WHERE orders_id = ".(int)$orders_id);
	}

	public function setPaycode($orders_id, $paycode, $amount = null) {
		$query = "REPLACE INTO paypal_paylink SET orders_id = ".(int)$orders_id.", paycode = '".xtc_db_input($paycode)."', amount = ".($amount === null ? 'null' : (double)$amount);
		xtc_db_query($query);
	}

	/**
	* returns TRUE if configured credentials are valid, FALSE otherwise
	*/
	public function checkAPIAccess() {
		$valid = false;
		$setECReqDetails = new SetExpressCheckoutRequestDetailsType();
		$setECReqType = new SetExpressCheckoutRequestType();
		$setECReqType->SetExpressCheckoutRequestDetails = $setECReqDetails;
		$setECReq = new SetExpressCheckoutReq();
		$setECReq->SetExpressCheckoutRequest = $setECReqType;

		$paypalService = new PayPalAPIInterfaceServiceService();

		$valid = false;
		try {
			$setECResponse = $paypalService->SetExpressCheckout($setECReq, $this->_getCredentials());
			if($setECResponse->Errors[0]->ErrorCode != 10002) {
				// ErrorCode 10002 = security error b/c of invalid credentials
				$valid = true;
			}
		} catch (Exception $ex) {
			//die('ERROR: '.print_r($ex, true));
			$this->_log('ERROR in '.__FUNCTION__.': '. $ex->getMessage());
			$valid = $ex->getMessage();
		}

		$t_valid_api = is_bool($valid) ? $valid : false;
		gm_set_conf('PAYPAL_VALID_API', strtoupper(var_export($t_valid_api, true)));

		return $valid;
	}

	public function prepareExpressCheckout() {
		$this->_log("preparing express checkout");
		$config = $this->getConfigArray();
		$cart = $_SESSION['cart'];
		// print_r($_SESSION); die();

		$return_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'checkout_paypal_prepare.php'.'?'.SID;
		$cancel_url = GM_HTTP_SERVER.DIR_WS_CATALOG.FILENAME_SHOPPING_CART.'?'.SID;
		$notify_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'paypal_ipn.php';
		$currencyCode = $_SESSION['currency'];
		$paymentDetails = new PaymentDetailsType();
		$paymentDetails->PaymentAction = $config['paymentaction']; // Sale|Authorization|Order
		if($_SESSION['customers_status']['customers_status_id'] == DEFAULT_CUSTOMERS_STATUS_ID_GUEST) {
			$address_override = '0';
		}
		else {
			$ab_entry = $this->getAddressBookEntry($_SESSION['customer_default_address_id']);
			$address = new AddressType();
			$address->CityName = $this->transcodeOutbound($ab_entry['entry_city']);
			$address->Name = $this->transcodeOutbound($ab_entry['entry_firstname'] .' '. $ab_entry['entry_lastname']);
			$address->Street1 = $this->transcodeOutbound($ab_entry['entry_street_address']);
			$country_data = $this->_findCountryByID($ab_entry['entry_country_id']);
			if($this->stateOrProvinceRequired($country_data['countries_iso_code_2'])) {
				if(!empty($_SESSION['paypal_state'])) {
					$address->StateOrProvince = $_SESSION['paypal_state'];
				}
				else {
					die('no state');
				}
			}
			$address->PostalCode = $ab_entry['entry_postcode'];
			$address->Country = $country_data['countries_iso_code_2'];
			#$address->Phone = $this->transcodeOutbound($order->customer['telephone']);
			$paymentDetails->ShipToAddress = $address;
			$address_override = '1';
		}
		// items
		$items = $cart->get_products();
		$itemTotal = 0;
		foreach($items as $item_no => $item) {
			$qty_is_non_int = abs($item['quantity'] - round($item['quantity'])) > 0;
			$has_qty_unit = !empty($item['unit_name']);
			$has_non_int_graduated_qty = $this->_getProductsGraduatedQty($item['id']) != 1;
			if($has_qty_unit || $has_non_int_graduated_qty) {
				// workaround for fractional quantities (which are unsupported by PayPal)
				$quantity = 1;
				$amount = $item['price'] * $item['quantity'];
				$unit_name = empty($item['unit_name']) ? 'x' : $item['unit_name'];
				$quantity_prefix = $item['quantity'] .' '.$unit_name.' ';
			}
			else {
				$quantity = $item['quantity'];
				$amount = $item['price'];
				$quantity_prefix = '';
			}
			if($amount < 0.01) {
				continue;
			}
			$itemDetails = new PaymentDetailsItemType();
			$itemDetails->Name = $this->transcodeOutbound($quantity_prefix . $item['name']);
			$itemDetails->Number = $this->transcodeOutbound($item['model']);
			if($_SESSION['customers_status']['customers_status_show_price_tax'] == 0)
			{
				$item_tax = $GLOBALS['xtPrice']->calcTax($amount, $GLOBALS['xtPrice']->TAX[$item['tax_class_id']]);
				$amount += $item_tax;
			}
			$itemAmount = new BasicAmountType($currencyCode, self::_formatAmount($amount));
			$itemDetails->Amount = $itemAmount;
			$itemDetails->Quantity = number_format((int)$quantity, 0, '', '');
			$item_category = $this->_productIsPhysical($item['id'], $item['attributes']) ? 'Physical' : 'Digital';
			$itemDetails->ItemCategory = $item_category;
			$paymentDetails->PaymentDetailsItem[$item_no] = $itemDetails;
			$itemTotal += ($amount * $quantity);
		}
		$paymentDetails->ItemTotal = new BasicAmountType($currencyCode, self::_formatAmount($itemTotal));
		$orderTotal = $itemTotal;
		if(isset($_SESSION['shipping'])) {
			$shippingTotal = new BasicAmountType($currencyCode, self::_formatAmount($_SESSION['shipping']['cost']));
			$paymentDetails->ShippingTotal = $shippingTotal;
			$orderTotal += $_SESSION['shipping']['cost'];
		}
		$paymentDetails->OrderTotal = new BasicAmountType($currencyCode, self::_formatAmount($orderTotal));
		$paymentDetails->NotifyURL = $notify_url;
		$setECReqDetails = new SetExpressCheckoutRequestDetailsType();
		$setECReqDetails->PaymentDetails[0] = $paymentDetails;
		$setECReqDetails->CancelURL = $cancel_url;
		$setECReqDetails->ReturnURL = $return_url;
		$setECReqDetails->PaymentAction = $config['paymentaction']; // Sale|Authorization|Order
		$setECReqDetails->AddressOverride = $address_override;

		// Shipping details
		$setECReqDetails->NoShipping = 2;
		$setECReqDetails->AddressOverride = 0;
		$setECReqDetails->ReqConfirmShipping = 0;
		$setECReqDetails->AllowNote = 1;

		// Display options
		$setECReqDetails->cppheaderimage = $config['cppheaderimage'];
		$setECReqDetails->cpplogoimage = $config['cpplogoimage'];
		$setECReqDetails->cppheaderbordercolor = $config['cppheaderbordercolor'];
		$setECReqDetails->cppheaderbackcolor = $config['cppheaderbackcolor'];
		$setECReqDetails->cpppayflowcolor = $config['cpppayflowcolor'];
		$setECReqDetails->cppcartbordercolor = $config['cppcartbordercolor'];
		$setECReqDetails->PageStyle = $config['pagestyle'];
		$setECReqDetails->BrandName = $config['brandname'];

		$setECReqType = new SetExpressCheckoutRequestType();
		$setECReqType->SetExpressCheckoutRequestDetails = $setECReqDetails;
		$setECReq = new SetExpressCheckoutReq();
		$setECReq->SetExpressCheckoutRequest = $setECReqType;
		$this->_log("setECReq:\n". print_r($setECReq, true));

		$paypalService = new PayPalAPIInterfaceServiceService();

		$token = false;

		try {
			$setECResponse = $paypalService->SetExpressCheckout($setECReq, $this->_getCredentials());
			$this->_log("setECResponse:\n". print_r($setECResponse, true));
			if(isset($setECResponse)) {
				if($setECResponse->Ack == 'Success') {
					$token = $setECResponse->Token;
					//$this->saveToken($orders_id, $token);
					//$this->saveCorrelationId($orders_id, (string)$setECResponse->CorrelationID);
				}
			}
		} catch (Exception $ex) {
			$this->_log("ERROR in SetExpressCheckout:\n".print_r($ex, true));
			# die('error occurred, see log');
		}

		return $token;
	}

	protected function _makePaymentDetailsItemArray($p_order)
	{
		$this->_log('p_order in '.__FUNCTION__."\n".print_r($p_order, true));
		$t_paymentdetailsitem = array();

		$itemTotal = 0;
		foreach($p_order->products as $item_no => $product) {
			$qty_is_non_int = abs($product['qty'] - round($product['qty'])) > 0;
			$has_qty_unit = !empty($product['unit_name']);
			$has_non_int_graduated_qty = $this->_getProductsGraduatedQty($product['id']) != 1;
			if($has_qty_unit || $has_non_int_graduated_qty) {
				// workaround for fractional quantities (which are unsupported by PayPal)
				$qty = 1;
				$amount = $product['final_price'];
				$unit_name = empty($product['unit_name']) ? 'x' : $product['unit_name'];
				$qty_prefix = $product['qty'] .' '.$unit_name.' ';
			}
			else {
				$qty = $product['qty'];
				$amount = $product['price'];
				$qty_prefix = '';
			}
			if($amount < 0.01) {
				continue;
			}
			$attributes = array();
			$attributes_suffix = '';
			if(false && !empty($product['attributes'])) {
				foreach($product['attributes'] as $attrib) {
					$attributes[] = $attrib['option'].': '.$attrib['value'];
				}
				$attributes_suffix = ' (' . implode(', ', $attributes) . ')';
			}
			$itemAmount = new BasicAmountType($currencyCode, self::_formatAmount($amount));
			$itemDetails = new PaymentDetailsItemType();
			$itemDetails->Name = $this->transcodeOutbound($qty_prefix . $product['name'] . $attributes_suffix);
			$itemDetails->Number = $this->transcodeOutbound($product['model']);
			$itemDetails->Amount = $itemAmount;
			$itemDetails->Quantity = number_format((int)$qty, 0, '', '');
			$item_category = $this->_productIsPhysical($product['id'], $product['attributes']) ? 'Physical' : 'Digital';
			$itemDetails->ItemCategory = $item_category;

			$t_paymentdetailsitem[$item_no] = $itemDetails;
			$itemTotal += ($amount * $qty);
		}
		$orderTotal = $p_order->info['pp_total'];
		$remainingTotal = $orderTotal - $p_order->info['pp_shipping'] - $itemTotal;
		if(abs($remainingTotal) < 0.01) {
			$remainingTotal = 0;
		}
		$this->_log("remaining total: $remainingTotal");
		if($remainingTotal > 0) {
			// additional fees
			$itemAmount = new BasicAmountType($currencyCode, self::_formatAmount($remainingTotal));
			$itemDetails = new PaymentDetailsItemType();
			$itemDetails->Name = $this->transcodeOutbound($this->get_text('fees_tax'));
			$itemDetails->Amount = $itemAmount;
			$itemDetails->Quantity = 1;
			$t_paymentdetailsitem[] = $itemDetails;
			$itemTotal += $remainingTotal;
		}
		else if($remainingTotal < 0) {
			// discount
			$itemAmount = new BasicAmountType($currencyCode, self::_formatAmount($remainingTotal));
			$itemDetails = new PaymentDetailsItemType();
			$itemDetails->Name = $this->transcodeOutbound($this->get_text('discount'));
			$itemDetails->Amount = $itemAmount;
			$itemDetails->Quantity = 1;
			$t_paymentdetailsitem[] = $itemDetails;
			$itemTotal += $remainingTotal;
		}

		$t_itemdata = array(
			'order_total' => $orderTotal,
			'item_total' => $itemTotal,
			'paymentdetailsitem' => $t_paymentdetailsitem,
		);
		return $t_itemdata;
	}

	public function prepareClassicCheckout($orders_id, $order) {
		$this->_log("preparing classic checkout for $orders_id");
		$config = $this->getConfigArray();
		$order = new order($orders_id);
		$totaldata = $order->getTotalData($orders_id);
		// do SetExpressCheckout
		$return_url = GM_HTTP_SERVER.DIR_WS_CATALOG.FILENAME_CHECKOUT_PROCESS.'?'.SID;
		$cancel_url = GM_HTTP_SERVER.DIR_WS_CATALOG.FILENAME_CHECKOUT_PAYMENT.'?'.SID;
		$notify_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'paypal_ipn.php';

		$currencyCode = $order->info['currency'];
		$shippingTotal = new BasicAmountType($currencyCode, self::_formatAmount($order->info['pp_shipping']));

		$paymentDetails = new PaymentDetailsType();

		$address = new AddressType();
		$address->CityName = $this->transcodeOutbound($order->delivery['city']);
		$address->Name = $this->transcodeOutbound($order->delivery['firstname'] .' '. $order->delivery['lastname']);
		$address->Street1 = $this->transcodeOutbound($order->delivery['street_address']);
		if($this->stateOrProvinceRequired($order->delivery['country_iso_2'])) {
			/*
			if(!empty($order->delivery['state'])) {
				$address->StateOrProvince = $order->delivery['state'];
			}
			else */
			if(!empty($_SESSION['paypal_state'])) {
				$address->StateOrProvince = $_SESSION['paypal_state'];
			}
			else {
				die('no state');
			}
		}
		$address->PostalCode = $order->delivery['postcode'];
		$address->Country = $order->delivery['country_iso_2'];
		$address->Phone = $this->transcodeOutbound($order->customer['telephone']);
		$paymentDetails->ShipToAddress = $address;
		$t_itemdata = $this->_makePaymentDetailsItemArray($order);
		$paymentDetails->PaymentDetailsItem = $t_itemdata['paymentdetailsitem'];

		$paymentDetails->ItemTotal = new BasicAmountType($currencyCode, self::_formatAmount($t_itemdata['item_total']));
		$paymentDetails->OrderTotal = new BasicAmountType($currencyCode, self::_formatAmount($t_itemdata['order_total']));
		$paymentDetails->PaymentAction = $config['paymentaction']; // Sale|Authorization|Order
		$paymentDetails->ShippingTotal = $shippingTotal;
		$paymentDetails->NotifyURL = $notify_url;

		$setECReqDetails = new SetExpressCheckoutRequestDetailsType();
		$setECReqDetails->InvoiceID = $orders_id;
		$setECReqDetails->PaymentDetails[0] = $paymentDetails;
		$setECReqDetails->CancelURL = $cancel_url;
		$setECReqDetails->ReturnURL = $return_url;
		$setECReqDetails->PaymentAction = $config['paymentaction']; // Sale|Authorization|Order
		
		// Shipping details
		$setECReqDetails->NoShipping = 1;
		$setECReqDetails->AddressOverride = 1;
		$setECReqDetails->ReqConfirmShipping = 0;

		// Display options
		$setECReqDetails->cppheaderimage = $config['cppheaderimage'];
		$setECReqDetails->cpplogoimage = $config['cpplogoimage'];
		$setECReqDetails->cppheaderbordercolor = $config['cppheaderbordercolor'];
		$setECReqDetails->cppheaderbackcolor = $config['cppheaderbackcolor'];
		$setECReqDetails->cpppayflowcolor = $config['cpppayflowcolor'];
		$setECReqDetails->cppcartbordercolor = $config['cppcartbordercolor'];
		$setECReqDetails->PageStyle = $config['pagestyle'];
		$setECReqDetails->BrandName = $config['brandname'];

		// Advanced options
		$setECReqDetails->AllowNote = 0;

		$setECReqType = new SetExpressCheckoutRequestType();
		$setECReqType->SetExpressCheckoutRequestDetails = $setECReqDetails;
		$setECReq = new SetExpressCheckoutReq();
		$setECReq->SetExpressCheckoutRequest = $setECReqType;
		$this->_log("setECReq:\n". print_r($setECReq, true));

		$paypalService = new PayPalAPIInterfaceServiceService();
		$token = false;
		try {
			$setECResponse = $paypalService->SetExpressCheckout($setECReq, $this->_getCredentials());
			$this->_log("setECResponse:\n". print_r($setECResponse, true));
			if(isset($setECResponse)) {
				if($setECResponse->Ack == 'Success') {
					$token = $setECResponse->Token;
					$this->saveToken($orders_id, $token);
					$this->saveCorrelationId($orders_id, (string)$setECResponse->CorrelationID);
					$this->savePaymentaction($orders_id);
				}
			}
		} catch (Exception $ex) {
			$this->_log("ERROR in SetExpressCheckout (classic checkout):\n".print_r($ex, true));
			#die('error occurred, see log');
		}

		return $token;
	}


	public function prepareIndependentPayment($orders_id, $amount = null) {
		$this->_log("preparing independent payment for $orders_id");
		$config = $this->getConfigArray();
		$order = new order($orders_id);
		$totaldata = $order->getTotalData($orders_id);
		// do SetExpressCheckout
		$return_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'paypal_payment.php'.'?'.SID;
		$cancel_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'paypal_payment.php'.'?'.SID;
		$notify_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'paypal_ipn.php';

		$currencyCode = $order->info['currency'];
		$shippingTotal = new BasicAmountType($currencyCode, self::_formatAmount($order->info['pp_shipping']));

		$paymentDetails = new PaymentDetailsType();

		$address = new AddressType();
		$address->CityName = $this->transcodeOutbound($order->delivery['city']);
		$address->Name = $this->transcodeOutbound($order->delivery['firstname'] .' '. $order->delivery['lastname']);
		$address->Street1 = $this->transcodeOutbound($order->delivery['street_address']);
		if($this->stateOrProvinceRequired($order->delivery['country_iso_2'])) {
			if(!empty($_SESSION['paypal_state'])) {
				$address->StateOrProvince = $_SESSION['paypal_state'];
			}
			else {
				die('no state');
			}
		}
		$address->PostalCode = $this->transcodeOutbound($order->delivery['postcode']);
		$address->Country = $order->delivery['country_iso_2'];
		$address->Phone = $this->transcodeOutbound($order->customer['telephone']);
		$paymentDetails->ShipToAddress = $address;

		$orderTotal = $amount !== null ? $amount : $order->info['pp_total'];

		$itemAmount = new BasicAmountType($currencyCode, self::_formatAmount($orderTotal));
		$itemDetails = new PaymentDetailsItemType();
		$itemDetails->Name = $this->transcodeOutbound($this->get_text('general_payment').' '.$orders_id);
		$itemDetails->Amount = $itemAmount;
		$itemDetails->Quantity = 1;
		$paymentDetails->PaymentDetailsItem[$item_no] = $itemDetails;

		$paymentDetails->ItemTotal = new BasicAmountType($currencyCode, self::_formatAmount($orderTotal));
		$paymentDetails->OrderTotal = new BasicAmountType($currencyCode, self::_formatAmount($orderTotal));
		$paymentDetails->PaymentAction = $config['paymentaction']; // Sale|Authorization|Order
		//$paymentDetails->ShippingTotal = $shippingTotal;
		$paymentDetails->NotifyURL = $notify_url;

		$setECReqDetails = new SetExpressCheckoutRequestDetailsType();
		$setECReqDetails->InvoiceID = $orders_id;
		$setECReqDetails->PaymentDetails[0] = $paymentDetails;
		$setECReqDetails->CancelURL = $cancel_url;
		$setECReqDetails->ReturnURL = $return_url;
		$setECReqDetails->PaymentAction = $config['paymentaction']; // Sale|Authorization|Order
		
		// Shipping details
		$setECReqDetails->NoShipping = 1;
		$setECReqDetails->AddressOverride = 1;
		$setECReqDetails->ReqConfirmShipping = 0;

		// Display options
		$setECReqDetails->cppheaderimage = $config['cppheaderimage'];
		$setECReqDetails->cpplogoimage = $config['cpplogoimage'];
		$setECReqDetails->cppheaderbordercolor = $config['cppheaderbordercolor'];
		$setECReqDetails->cppheaderbackcolor = $config['cppheaderbackcolor'];
		$setECReqDetails->cpppayflowcolor = $config['cpppayflowcolor'];
		$setECReqDetails->cppcartbordercolor = $config['cppcartbordercolor'];
		$setECReqDetails->PageStyle = $config['pagestyle'];
		$setECReqDetails->BrandName = $config['brandname'];

		// Advanced options
		$setECReqDetails->AllowNote = 0;

		$setECReqType = new SetExpressCheckoutRequestType();
		$setECReqType->SetExpressCheckoutRequestDetails = $setECReqDetails;
		$setECReq = new SetExpressCheckoutReq();
		$setECReq->SetExpressCheckoutRequest = $setECReqType;
		$this->_log("setECReq:\n". print_r($setECReq, true));

		$paypalService = new PayPalAPIInterfaceServiceService();

		$token = false;

		try {
			$setECResponse = $paypalService->SetExpressCheckout($setECReq, $this->_getCredentials());
			$this->_log("setECResponse:\n". print_r($setECResponse, true));
		} catch (Exception $ex) {
			$this->_log("ERROR in SetExpressCheckout (independent payment)\n".print_r($ex, true));
			# die('error occurred, see log');
		}

		if(isset($setECResponse) && $setECResponse->Ack == 'Success') {
			$token = $setECResponse->Token;
			/*
			$this->saveToken($orders_id, $token);
			$this->saveCorrelationId($orders_id, (string)$setECResponse->CorrelationID);
			$this->savePaymentaction($orders_id);
			*/
		}
		else {
			throw new GMPayPalNoTokenException((string)$setECResponse->Errors[0]->LongMessage);
		}

		return $token;
	}

	public function getRedirectURL($cmd, $token, $p_useraction = null) {
		$pp_config = PPConfigManager::getInstance();
		$base_url = $pp_config->get('service.RedirectURL');
		$url = $base_url . $cmd . '&token=' . $token;
		if($p_useraction !== null)
		{
			$url .= '&useraction='.$p_useraction;
		}
		return $url;
	}

	public function saveToken($orders_id, $token) {
		$query = "INSERT INTO orders_paypal (orders_id, token) VALUES (:orders_id, ':token') ON DUPLICATE KEY UPDATE token = ':token'";
		$query = strtr($query, array(':orders_id' => (int)$orders_id, ':token' => xtc_db_input($token)));
		xtc_db_query($query);
	}

	public function getToken($orders_id) {
		$data = $this->getOrderData($orders_id);
		return $data['token'];
	}

	public function savePaymentaction($orders_id, $paymentaction = null) {
		if($paymentaction === null) {
			$config = $this->getConfigArray();
			$paymentaction = $config['paymentaction'];
		}
		$query = "INSERT INTO orders_paypal (orders_id, paymentaction) VALUES (:orders_id, ':paymentaction') ON DUPLICATE KEY UPDATE paymentaction = ':paymentaction'";
		$query = strtr($query, array(':orders_id' => (int)$orders_id, ':paymentaction' => xtc_db_input($paymentaction)));
		xtc_db_query($query);
	}

	public function getPaymentaction($orders_id) {
		$data = $this->getOrderData($orders_id);
		return $data['paymentaction'];
	}

	public function saveCorrelationId($orders_id, $correlation_id) {
		$query = "INSERT INTO orders_paypal (orders_id, correlation_id) VALUES (:orders_id, ':correlation_id') ON DUPLICATE KEY UPDATE correlation_id = ':correlation_id'";
		$query = strtr($query, array(':orders_id' => (int)$orders_id, ':correlation_id' => $correlation_id));
		xtc_db_query($query);
	}

	public function savePayerId($orders_id, $payer_id) {
		$query = "INSERT INTO orders_paypal (orders_id, payer_id) VALUES (:orders_id, ':payer_id') ON DUPLICATE KEY UPDATE payer_id = ':payer_id'";
		$query = strtr($query, array(':orders_id' => (int)$orders_id, ':payer_id' => xtc_db_input($payer_id)));
		//$this->_log("savePayerId query: $query");
		xtc_db_query($query);
	}

	public function getOrderData($orders_id) {
		$query = "SELECT * FROM orders_paypal WHERE orders_id = :orders_id";
		$query = strtr($query, array(':orders_id' => (int)$orders_id));
		$result = xtc_db_query($query, 'db_link', false);
		$data = false;
		while($row = xtc_db_fetch_array($result)) {
			$data = $row;
		}
		if($data === false) {
			throw new GMPayPalNoDataException();
		}
		else {
			return $data;
		}
	}

	protected function _getProductsGraduatedQty($products_id) {
		$query = "SELECT gm_graduated_qty FROM products WHERE products_id = ".(int)$products_id;
		$result = xtc_db_query($query);
		$gm_graduated_qty = 1;
		while($row = xtc_db_fetch_array($result)) {
			$gm_graduated_qty = $row['gm_graduated_qty'];
		}
		return $gm_graduated_qty;
	}


	public function getExpressCheckoutDetails($orders_id, $force_retrieve = false) {
		$details = false;

		// check local data
		$query = "SELECT * FROM paypal_checkoutdetails WHERE orders_id = :orders_id";
		$query = strtr($query, array(':orders_id' => (int)$orders_id));
		$result = xtc_db_query($query);
		if(xtc_db_num_rows($result) > 0) {
			$row = xtc_db_fetch_array($result);
			$details = array(
				'source' => 'db',
				'retrievaltime' => $row['retrievaltime'],
				'checkoutdetails' => unserialize($row['checkoutdetails']),
			);
		}

		if($details === false || $force_retrieve === true) {
			$token = $this->getToken($orders_id);
			$getECResponse = $this->getExpressCheckoutDetailsByToken($token);
			if($getECResponse !== false)
			{
				if(empty($getECResponse->Errors) === false)
				{
					foreach($getECResponse->Errors as $error)
					{
						if((string)$error->ErrorCode == '10411')
						{
							throw new GMPayPalTokenExpiredException();
						}
					}
				}
				$retrievaltime = $this->saveGetExpressCheckoutDetails($orders_id, $getECResponse);
				$this->_log("retrieved CheckoutDetails for $orders_id");
				$this->_log(print_r($getECResponse, true));
				$details = array(
					'source' => 'online',
					'retrievaltime' => $retrievaltime,
					'checkoutdetails' => $getECResponse,
				);
			}
			else
			{
				throw new GMPayPalNoCheckoutDetailsException();
			}
		}
		return $details;
	}

	public function saveGetExpressCheckoutDetails($orders_id, $getECResponse) {
		$retrievaltime = date('Y-m-d H:i:s');
		$save_query = "REPLACE INTO paypal_checkoutdetails (orders_id, retrievaltime, checkoutdetails) VALUES (:orders_id, ':retrievaltime', ':checkoutdetails')";
		$save_query = strtr($save_query, array(':orders_id' => (int)$orders_id, ':retrievaltime' => $retrievaltime, ':checkoutdetails' => xtc_db_input(serialize($getECResponse))));
		xtc_db_query($save_query);
		return $retrievaltime;		
	}

	public function getExpressCheckoutDetailsByToken($token) {
		$this->_log('retrieving CheckoutDetails for token'. $token);
		$getExpressCheckoutDetailsRequest = new GetExpressCheckoutDetailsRequestType($token);
		$getExpressCheckoutReq = new GetExpressCheckoutDetailsReq();
		$getExpressCheckoutReq->GetExpressCheckoutDetailsRequest = $getExpressCheckoutDetailsRequest;
		$paypalService = new PayPalAPIInterfaceServiceService();
		try {
			$getECResponse = $paypalService->GetExpressCheckoutDetails($getExpressCheckoutReq, $this->_getCredentials());
		}
		catch(Exception $ex) {
			$this->_log('ERROR retrieving CheckoutDetails for token '.$token.': '.$ex->getMessage());
			$getECResponse = false;
		}
		return $getECResponse;
	}

	public function doExpressCheckoutPayment($orders_id, $orderdata = null, $amount = null) {
		$config = $this->getConfigArray();
		$order = new order($orders_id);
		$orderdata = $orderdata === null ? $this->getOrderData($orders_id) : $orderdata;
		$amount = $amount === null ? $order->info['pp_total'] : $amount;
		$paymentAction = $config['paymentaction']; //$orderdata['paymentaction']; // Sale|Authorization|Order
		$notify_url = GM_HTTP_SERVER.DIR_WS_CATALOG.'paypal_ipn.php';

		$orderTotal = new BasicAmountType();
		$orderTotal->currencyID = $order->info['currency'];
		$orderTotal->value = self::_formatAmount($amount);

		$PaymentDetails = new PaymentDetailsType();
		$PaymentDetails->InvoiceID = $orders_id;
		$PaymentDetails->OrderTotal = $orderTotal;
		$PaymentDetails->ShippingTotal = new BasicAmountType($currencyCode, self::_formatAmount($order->info['pp_shipping']));
		$PaymentDetails->NotifyURL = $notify_url;
		$t_itemdata = $this->_makePaymentDetailsItemArray($order);
		$PaymentDetails->PaymentDetailsItem = $t_itemdata['paymentdetailsitem'];
		$PaymentDetails->ItemTotal = new BasicAmountType($currencyCode, self::_formatAmount($t_itemdata['item_total']));
		if(strpos($order->info['shipping_class'], 'selfpickup') === false) {
			// address MUST NOT be set for selfpickup (not eligible for seller protection)
			$address = new AddressType();
			$address->CityName = $this->transcodeOutbound($order->delivery['city']);
			$address->Name = $this->transcodeOutbound($order->delivery['firstname'] .' '. $order->delivery['lastname']);
			$address->Street1 = $this->transcodeOutbound($order->delivery['street_address']);
			if($this->stateOrProvinceRequired($order->delivery['country_iso_2'])) {
				/*
				if(!empty($order->delivery['state'])) {
					$address->StateOrProvince = $order->delivery['state'];
				}
				else */
				if(!empty($_SESSION['paypal_state'])) {
					$address->StateOrProvince = $_SESSION['paypal_state'];
				}
				else {
					die('no state');
				}
			}
			$address->PostalCode = $this->transcodeOutbound($order->delivery['postcode']);
			$address->Country = $order->delivery['country_iso_2'];
			$address->Phone = $this->transcodeOutbound($order->customer['telephone']);
			$PaymentDetails->ShipToAddress = $address;
		}

		$DoECRequestDetails = new DoExpressCheckoutPaymentRequestDetailsType();
		$DoECRequestDetails->PayerID = $orderdata['payer_id'];
		$DoECRequestDetails->Token = $orderdata['token'];
		$DoECRequestDetails->PaymentAction = $paymentAction;
		$DoECRequestDetails->PaymentDetails[0] = $PaymentDetails;

		$DoECRequest = new DoExpressCheckoutPaymentRequestType();
		$DoECRequest->DoExpressCheckoutPaymentRequestDetails = $DoECRequestDetails;

		$DoECReq = new DoExpressCheckoutPaymentReq();
		$DoECReq->DoExpressCheckoutPaymentRequest = $DoECRequest;

		$paypalService = new PayPalAPIInterfaceServiceService();
		$success = false;
		$doECResponse = null;
		try {
			$this->_log("doExpressCheckout Request:\n".print_r($DoECReq, true));
			$doECResponse = $paypalService->DoExpressCheckoutPayment($DoECReq, $this->_getCredentials());
			$this->_log("doECPayment for $orders_id: ". (string)$doECResponse->Ack ." paymentAction = ".$paymentAction);
			$this->saveTransaction($orders_id, $doECResponse);
			$success = (string)$doECResponse->Ack == 'Success';
		}
		catch(Exception $ex) {
			$this->_log('ERROR performing doExpressCheckoutPayment for $orders_id: '.$ex->getMessage());
		}

		$result = array(
			'success' => $success,
			'doecresponse' => $doECResponse,
		);

		return $result;
	}

	public function saveTransaction($orders_id, $doECResponse) {
		$this->_log("saveTransaction doECResponse:\n".print_r($doECResponse, true));
		$config = $this->getConfigArray();
		$paymentAction = $config['paymentaction'];
		$pp_transaction = array(
			'orders_id' => $orders_id,
			'correlation_id' => (string)$doECResponse->CorrelationID,
			'ack' => (string)$doECResponse->Ack,
			'timestamp' => (string)$doECResponse->Timestamp,
			'transaction' => 'doexpresscheckoutpayment',
			'version_build' => (string)$doECResponse->Version .' '. (string)$doECResponse->Build,
			'transaction_id' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->TransactionID,
			'parenttransaction_id' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->ParentTransactionID,
			'transaction_type' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->TransactionType .' - '. $paymentAction,
			'payment_type' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->PaymentType,
			'payment_date' => (string)(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->PaymentDate,
			'grossamount' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->GrossAmount->value .' '.
				(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->GrossAmount->currencyID,
			'feeamount' => (empty($doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->FeeAmount) ? '' :
				(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->FeeAmount->value .' '.
				(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->FeeAmount->currencyID),
			'settleamount' => (empty($doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->SettleAmount) ? '' :
				(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->SettleAmount->value .' '.
				(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->SettleAmount->currencyID),
			'taxamount'=> (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->TaxAmount->value .' '.
				(string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->TaxAmount->currencyID,
			'exchange_rate' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->ExchangeRate,
			'paymentstatus' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->PaymentStatus,
			'expected_echeck_clear_date' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->ExpectedeCheckClearDate,
			'pending_reason' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->PendingReason,
			'reason_code' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->ReasonCode,
			'hold_decision' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->HoldDecision,
			'protection_eligibility' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->ProtectionEligibility,
			'protection_eligibility_type' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->ProtectionEligibilityType,
			'secure_merchant_account_id' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->SellerDetails->SecureMerchantAccountID,
			'payment_error' => (string)$doECResponse->DoExpressCheckoutPaymentResponseDetails->PaymentInfo[0]->PaymentError,
			'raw_response' => serialize($doECResponse),
			'success_page_redirect_requested' => ($doECResponse->DoExpressCheckoutPaymentResponseDetails->SuccessPageRedirectRequested ? '1' : '0'),
			'errors' => serialize($doECResponse->Errors),
		);
		xtc_db_perform('paypal_transactions', $pp_transaction, 'insert');
	}

	public function saveCaptureTransaction($orders_id, $doCaptureResponse) {
		$this->_log("saveCaptureTransaction doCaptureResponse:\n".print_r($doCaptureResponse, true));
		$pp_transaction = array(
			'orders_id' => $orders_id,
			'correlation_id' => (string)$doCaptureResponse->CorrelationID,
			'ack' => (string)$doCaptureResponse->Ack,
			'timestamp' => (string)$doCaptureResponse->Timestamp,
			'transaction' => 'docapture',
			'version_build' => (string)$doCaptureResponse->Version .' '. (string)$doCaptureResponse->Build,
			'transaction_id' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->TransactionID,
			'parenttransaction_id' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->ParentTransactionID,
			'transaction_type' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->TransactionType,
			'payment_type' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->PaymentType,
			'payment_date' => (string)(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->PaymentDate,
			'grossamount' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->GrossAmount->value .' '.
				(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->GrossAmount->currencyID,
			'feeamount' => (empty($doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->FeeAmount) ? '' :
				(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->FeeAmount->value .' '.
				(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->FeeAmount->currencyID),
			'settleamount' => (empty($doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->SettleAmount) ? '' :
				(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->SettleAmount->value .' '.
				(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->SettleAmount->currencyID),
			'taxamount'=> (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->TaxAmount->value .' '.
				(string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->TaxAmount->currencyID,
			'exchange_rate' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->ExchangeRate,
			'paymentstatus' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->PaymentStatus,
			'expected_echeck_clear_date' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->ExpectedeCheckClearDate,
			'pending_reason' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->PendingReason,
			'reason_code' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->ReasonCode,
			'hold_decision' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->HoldDecision,
			'protection_eligibility' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->ProtectionEligibility,
			'protection_eligibility_type' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->ProtectionEligibilityType,
			'secure_merchant_account_id' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->SellerDetails->SecureMerchantAccountID,
			'payment_error' => (string)$doCaptureResponse->DoCaptureResponseDetails->PaymentInfo->PaymentError,
			'raw_response' => serialize($doCaptureResponse),
			'success_page_redirect_requested' => '0',
			'errors' => serialize($doCaptureResponse->Errors),
		);
		xtc_db_perform('paypal_transactions', $pp_transaction, 'insert');
	}

	public function saveAuthTransaction($orders_id, $doAuthResponse) {
		$this->_log("saveAuthTransaction doAuthResponse:\n".print_r($doAuthResponse, true));
		$pp_transaction = array(
			'orders_id' => (int)$orders_id,
			'correlation_id' => (string)$doAuthResponse->CorrelationID,
			'ack' => (string)$doAuthResponse->Ack,
			'timestamp' => (string)$doAuthResponse->Timestamp,
			'transaction' => 'doauthorize',
			'version_build' => (string)$doAuthResponse->Version .' '. (string)$doAuthResponse->Build,
			'transaction_id' => (string)$doAuthResponse->TransactionID,
			'grossamount' => (string)$doAuthResponse->Amount->value .' '. (string)$doAuthResponse->Amount->currencyID,
			'paymentstatus' => (string)$doAuthResponse->AuthorizationInfo->PaymentStatus,
			'pending_reason' => (string)$doAuthResponse->AuthorizationInfo->PendingReason,
			'protection_eligibility' => (string)$doAuthResponse->AuthorizationInfo->ProtectionEligibility,
			'protection_eligibility_type' => (string)$doAuthResponse->AuthorizationInfo->ProtectionEligibilityType,
			'raw_response' => serialize($doAuthResponse),
			'errors' => serialize($doAuthResponse->Errors),
		);
		xtc_db_perform('paypal_transactions', $pp_transaction, 'insert');
	}

	public function getTransactions($orders_id) {
		$query = "SELECT * FROM paypal_transactions WHERE orders_id = :orders_id ORDER BY ecresponse_id";
		$query = strtr($query, array(':orders_id' => (int)$orders_id));
		$result = xtc_db_query($query);
		$transactions = array();
		while($row = xtc_db_fetch_array($result)) {
			$transactions[] = $row;
		}
		return $transactions;
	}

	public function getNotificationsForTransaction($txn_id) {
		$query = "SELECT * FROM paypal_ipn WHERE txn_id = ':txn_id' ORDER BY paypal_ipn_id ASC";
		$query = strtr($query, array(':txn_id' => xtc_db_input($txn_id)));
		$result = xtc_db_query($query);
		$ipns = array();
		while($row = xtc_db_fetch_array($result)) {
			$ipns[] = $row;
		}
		return $ipns;
	}

	/*
	public function getRedirectURL($token) {
		if($config['sandbox_mode'] == 1) {
			$redirect_url = 'https://www.sandbox.paypal.com/cgi-bin/webscr?cmd=_complete-express-checkout&token=';
		}
		else {
			$redirect_url = 'https://www.paypal.com/cgi-bin/webscr?cmd=_complete-express-checkout&token=';	
		}
		$redirect_url .= $token;
		return $redirect_url;
	}
	*/

	public function getTransactionDetails($transaction_id, $update = false) {
		$transactionDetails = new GetTransactionDetailsRequestType();
		$transactionDetails->TransactionID = $transaction_id;
		$request = new GetTransactionDetailsReq();
		$request->GetTransactionDetailsRequest = $transactionDetails;
		$paypalService = new PayPalAPIInterfaceServiceService();
		try {
			$transDetailsResponse = $paypalService->GetTransactionDetails($request, $this->_getCredentials());
			$this->_log("TransactionDetails for $transaction_id:\n".print_r($transDetailsResponse, true));
		}
		catch (Exception $ex) {
			$this->_log('ERROR in getTransactionDetails: '.$ex->getMessage());
			throw $ex;
		}

		// update stored data
		if($update == true) {
			$paymentstatus = (string)$transDetailsResponse->PaymentTransactionDetails->PaymentInfo->PaymentStatus;
			$pending_reason = (string)$transDetailsResponse->PaymentTransactionDetails->PaymentInfo->PendingReason;
			$query = "UPDATE `paypal_transactions` SET `paymentstatus` = ':paymentstatus', `pending_reason` = ':pending_reason' WHERE `transaction_id` = ':transaction_id'";
			$query = strtr($query, array(':paymentstatus' => $paymentstatus, ':pending_reason' => $pending_reason, ':transaction_id' => $transaction_id));
			xtc_db_query($query);
		}

		return $transDetailsResponse;
	}

	public function capturePayment($orders_id, $transaction_id, $amount, $currency_code, $partial = false) {
		$amount = (double)$amount;
		$basic_amount = new BasicAmountType($currency_code, $amount);
		if($partial == true) {
			$complete_code = 'Complete';
		}
		else {
			$complete_code = 'NotComplete';
		}		
		$doCaptureRequest = new DoCaptureRequestType($transaction_id, $basic_amount, $complete_code);
		$doCaptureReq = new DoCaptureReq();
		$doCaptureReq->DoCaptureRequest = $doCaptureRequest;
		$paypalService = new PayPalAPIInterfaceServiceService();
		$this->_log("Capturing payment of $amount $currency_code for transaction $transaction_id");
		try {
			$doCaptureResponse = $paypalService->DoCapture($doCaptureReq, $this->_getCredentials());
			$this->saveCaptureTransaction($orders_id, $doCaptureResponse);
		}
		catch (Exception $ex) {
			$this->_log('ERROR in capturePayment: '.$ex->getMessage());
			throw $ex;
		}
		//$this->_log(print_r($doCaptureResponse, true));
		if(!empty($doCaptureResponse->Errors)) {
			$error_msg = array();
			foreach($doCaptureResponse->Errors as $error) {
				$error_msg[] = $error->ShortMessage .' - '. $error->LongMessage;
			}
			$error_msg_text = implode(', ', $error_msg);
			$this->_log("ERROR capturing payment: ".$error_msg_text);
			throw new Exception($error_msg_text);
		}
		else {
			$this->_log("Captured successfully");
		}
		return $doCaptureResponse;
	}

	public function authorizePayment($orders_id, $transaction_id, $amount, $currency_code) {
		$basic_amount = new BasicAmountType($currency_code, $amount);
		$doAuthRequest = new DoAuthorizationRequestType($transaction_id, $basic_amount);
		$doAuthReq = new DoAuthorizationReq();
		$doAuthReq->DoAuthorizationRequest = $doAuthRequest;
		$paypalService = new PayPalAPIInterfaceServiceService();
		try {
			$doAuthResponse = $paypalService->DoAuthorization($doAuthReq, $this->_getCredentials());
		} catch (Exception $ex) {
			$this->_log('ERROR in authorizePayment: '.$ex->getMessage());
			throw $ex;
		}
		if(!empty($doAuthResponse->Errors)) {
			$error_msg = array();
			foreach($doAuthResponse->Errors as $error) {
				$error_msg[] = $error->ShortMessage .' - '. $error->LongMessage;
			}
			$error_msg_text = implode(', ', $error_msg);
			$this->_log("ERROR authorizing payment: ".$error_msg_text);
			throw new Exception($error_msg_text);
		}
		else {
			$this->saveAuthTransaction($orders_id, $doAuthResponse);
			$this->_log("payment authorized: $transaction_id, $amount $currency_code");
			$this->_log(print_r($doAuthResponse, true));
		}
		return $doAuthResponse;
	}

	public function voidAuthorization($transaction_id) {
		$doVoidRequest = new DoVoidRequestType();
		$doVoidRequest->AuthorizationID = $transaction_id;
		$doVoidReq = new DoVoidReq();
		$doVoidReq->DoVoidRequest =$doVoidRequest;
		$paypalService = new PayPalAPIInterfaceServiceService();
		try {
			$doVoidResponse = $paypalService->DoVoid($doVoidReq, $this->_getCredentials());
			$this->_log("voidResponse:\n".print_r($doVoidResponse, true));
		}
		catch(Exception $e) {
			$this->_log('ERROR in voidAuthorization: '.$ex->getMessage());
			throw $ex;
		}
		return $doVoidResponse;
	}

	public function getNumberOfAuthorizations($orders_id) {
		$num_auth = $this->_countTransactions($orders_id, 'doauthorize');
		return $num_auth;
	}

	public function getNumberOfCaptures($orders_id) {
		$num_captures = $this->_countTransactions($orders_id, 'docapture');
		return $num_captures;
	}

	protected function _countTransactions($orders_id, $transaction) {
		$query = "SELECT COUNT(*) AS num FROM `paypal_transactions` WHERE `orders_id` = ".(int)$orders_id." AND `transaction` = '".$transaction."'";
		$num = 0;
		$result = xtc_db_query($query);
		while($row = xtc_db_fetch_array($result)) {
			$num = $row['num'];
		}
		return $num;
	}


	public function reverseTransaction($transaction_id) {
		$reverseTransactionRequestDetails = new ReverseTransactionRequestDetailsType();
		$reverseTransactionRequestDetails->TransactionID = $transaction_id;
		$reverseTransactionRequest = new ReverseTransactionRequestType();
		$reverseTransactionRequest->ReverseTransactionRequestDetails = $reverseTransactionRequestDetails;
		$reverseTransactionReq = new ReverseTransactionReq();
		$reverseTransactionReq->ReverseTransactionRequest = $reverseTransactionRequest;
		$paypalService = new PayPalAPIInterfaceServiceService();
		try {
			$reverseTransactionResponse = $paypalService->ReverseTransaction($reverseTransactionReq);
		} catch (Exception $ex) {
			$this->_log('ERROR in reverseTransaction: '.$ex->getMessage());
			throw $ex;
		}
		if(!empty($reverseTransactionResponse->Errors)) {
			$error_msg = array();
			foreach($reverseTransactionResponse->Errors as $error) {
				$error_msg[] = $error->ShortMessage .' - '. $error->LongMessage;
			}
			$error_msg_text = implode(', ', $error_msg);
			$this->_log("ERROR reversing transaction: ".$error_msg_text);
			throw new Exception($error_msg_text);
		}
		else {
			$this->_log("transaction reversed: $transaction_id");
			$this->_log(print_r($reverseTransactionResponse, true));
		}
		return $reverseTransactionResponse;
	}

	public function refund($orders_id, $transaction_id, $memo = '', $refund_type = 'Full', $amount = 0, $currency_code = '', $refund_source = 'default', $retry_until = '') {
		$this->_log("Performing refund on order $orders_id with transaction id $transaction_id type $refund_type amount $amount $currency_code from $refund_source");
		$refundRequest = new RefundTransactionRequestType();

		if($refund_type != "Full") {
			$refundRequest->Amount = new BasicAmountType($currency_code, $amount);
		}
		$refundRequest->RefundType = $refund_type; // Other|Full|Partial|ExternalDispute
		$refundRequest->TransactionID = $transaction_id;
		$refundRequest->RefundSource = $refund_source; // any|default|instant|eCheck
		$refundRequest->Memo = $memo;
		//$refundRequest->RetryUntil = $retry_until;

		$refundReq = new RefundTransactionReq();
		$refundReq->RefundTransactionRequest = $refundRequest;
		$this->_log("RefundTransactionRequest:\n".print_r($refundReq, true));

		$paypalService = new PayPalAPIInterfaceServiceService();
		try {
			$refundResponse = $paypalService->RefundTransaction($refundReq, $this->_getCredentials());
		} catch (Exception $ex) {
			$this->_log('ERROR in refund: '.$ex->getMessage());
			throw $ex;
		}
		if(!empty($refundResponse->Errors)) {
			$error_msg = array();
			foreach($refundResponse->Errors as $error) {
				$error_msg[] = $error->ShortMessage .' - '. $error->LongMessage;
			}
			$error_msg_text = implode(', ', $error_msg);
			$this->_log("ERROR refunding: ".$error_msg_text);
			throw new Exception($error_msg_text);
		}
		else {
			$log_data = array(
				'orders_id' => $orders_id, 
				'refundtransactionid' => (string)$refundResponse->RefundTransactionID,
				'netrefundvalue' => (string)$refundResponse->NetRefundAmount->value,
				'netrefundcurrency' => (string)$refundResponse->NetRefundAmount->currencyID,
				'feerefundvalue' => (string)$refundResponse->FeeRefundAmount->value,
				'feerefundcurrency' => (string)$refundResponse->FeeRefundAmount->currencyID,
				'grossrefundvalue' => (string)$refundResponse->GrossRefundAmount->value,
				'grossrefundcurrency' => (string)$refundResponse->GrossRefundAmount->currencyID,
				'totalrefundedvalue' => (string)$refundResponse->TotalRefundedAmount->value,
				'totalrefundedcurrency' => (string)$refundResponse->TotalRefundedAmount->currencyID,
				'refundinfostatus' => (string)$refundResponse->RefundInfo->RefundStatus,
				'refundinfopendingreason' => (string)$refundResponse->RefundInfo->PendingReason,
				'receiptdata' => (string)$refundResponse->ReceiptData,
				'msgsubid' => (string)$refundResponse->MsgSubID,
				'timestamp' => date('Y-m-d H:i:s', strtotime((string)$refundResponse->Timestamp)),
				'ack' => (string)$refundResponse->Ack,
				'correlationid' => (string)$refundResponse->CorrelationID,
				'errors' => (string)$refundResponse->Errors,
				'version' => (string)$refundResponse->Version,
				'build' => (string)$refundResponse->Build,
			);
			xtc_db_perform('paypal_refunds', $log_data, 'insert');
			$this->_log("refunded: $transaction_id");
			$this->_log(print_r($refundResponse, true));
		}
		return $refundResponse;
	}

	public function getRefunds($orders_id) {
		$query = "SELECT * FROM `paypal_refunds` WHERE orders_id = :orders_id ORDER BY `timestamp` ASC";
		$query = strtr($query, array(':orders_id' => (int)$orders_id));
		$result = xtc_db_query($query);
		$refunds = array();
		while($row = xtc_db_fetch_array($result)) {
			$refunds[] = $row;
		}
		return $refunds;
	}


	/* === auxilliary functions === */
	protected function _localEncodingIsLatin1() {
		$is_latin1 = strpos(strtolower($_SESSION['language_charset']), 'iso-8859-1') !== false;
		return $is_latin1;
	}

	/** transcode incoming string from UTF-8 to whatever the shop system is currenty using */
	public function transcodeInbound($string) {
		if($this->_localEncodingIsLatin1()) {
			$output = utf8_decode($string);
		}
		else {
			$output = $string;
		}
		return $output;
	}

	/** transcode incoming string from whatever the shop system is currenty using to UTF-8 */
	public function transcodeOutbound($string) {
		if($this->_localEncodingIsLatin1()) {
			$output = utf8_encode($string);
		}
		else {
			$output = $string;
		}
		return $output;
	}

	public function getAddressBookEntry($ab_id) {
		$query = "SELECT * FROM `address_book` WHERE `address_book_id` = ".(int)$ab_id;
		$result = xtc_db_query($query, 'db_link', false);
		$ab_entry = false;
		while($row = xtc_db_fetch_array($result)) {
			$ab_entry = $row;
		}
		return $ab_entry;
	}

	public function findOrCreateAddressBookEntry($customer_id, $address) {
		// heuristically split name by last occurrence of a space character
		if(preg_match('/(.*) (.*)/', $address['name'], $matches) == 1) {
			$firstname = $matches[1];
			$lastname = $matches[2];
		}
		else {
			$firstname = '';
			$lastname = $address['name'];
		}
		$country_id = $this->_findCountryByIso2($address['country']);
		$zone_id = $this->_findZoneByName($address['stateorprovince'], $country_id);

		$sendto_data = array( // a row in address_book
			'customers_id' => $customer_id,
			'entry_gender' => 'm',
			'entry_company' => '',
			'entry_firstname' => $firstname,
			'entry_lastname' => $lastname,
			'entry_street_address' => $address['street1'],
			'entry_suburb' => '',
			'entry_postcode' => $address['postalcode'],
			'entry_city' => $address['cityname'],
			'entry_state' => $address['stateorprovince'],
			'entry_country_id' => $country_id,
			'entry_zone_id' => $zone_id,
			'address_date_added' => 'now()',
			'address_last_modified' => 'now()',
			'address_class' => 'paypal',
		);

		$find_address_query  = "SELECT address_book_id, CONCAT(entry_firstname, ' ', entry_lastname) AS name FROM address_book ";
		$find_address_query .= "WHERE customers_id = :customers_id AND entry_street_address LIKE ':street_address' ";
		$find_address_query .= "AND entry_postcode LIKE ':postcode' AND entry_city LIKE ':city' "; // AND entry_state LIKE ':state' ";
		$find_address_query .= "AND entry_country_id = :country_id ";
		$find_address_query .= "HAVING name LIKE ':name' ";
		$find_address_query = strtr($find_address_query, array(
			':customers_id' => $customer_id,
			':street_address' => xtc_db_input($address['street1']),
			':postcode' => xtc_db_input($address['postalcode']),
			':city' => xtc_db_input($address['cityname']),
			':state' => xtc_db_input($address['stateorprovince']),
			':country_id' => $country_id,
			':name' => xtc_db_input($address['name']),
		));
		$ab_id = false;
		$this->_log('find_address_query: '.$find_address_query);
		$find_address_result = xtc_db_query($find_address_query, 'db_link', false);
		while($find_address_row = xtc_db_fetch_array($find_address_result)) {
			$ab_id = $find_address_row['address_book_id'];
		}
		if($ab_id !== false) {
			return $ab_id;
		}
		else {
			xtc_db_perform('address_book', $sendto_data, 'insert');
			$ab_id = xtc_db_insert_id();
			return $ab_id;
		}
		return false; // this should never happen
	}

	protected function _findCountryByIso2($iso2) {
		$query = "SELECT countries_id FROM countries WHERE countries_iso_code_2 = ':iso2'";
		$query = strtr($query, array(':iso2' => xtc_db_input($iso2)));
		$result = xtc_db_query($query);
		$countries_id = 0;
		while($row = xtc_db_fetch_array($result)) {
			$countries_id = $row['countries_id'];
		}
		return $countries_id;
	}

	protected function _findCountryByID($country_id) {
		$query = "SELECT * FROM `countries` WHERE countries_id = ':countries_id'";
		$query = strtr($query, array(':countries_id' => (int)$country_id));
		$result = xtc_db_query($query);
		$country_data = false;
		while($row = xtc_db_fetch_array($result)) {
			$country_data = $row;
		}
		return $country_data;
	}

	protected function _findZoneByName($name, $country_id = null) {
		$zone_id = 0;
		$query = "SELECT zone_id FROM zones WHERE zone_name LIKE ':name'";
		if(!empty($country_id)) {
			$query .= " AND zone_country_id = :country_id";
		}
		$query = strtr($query, array(':name' => xtc_db_input($name), ':country_id' => $country_id));
		$result = xtc_db_query($query);
		while($row = xtc_db_fetch_array($result)) {
			$zone_id = $row['zone_id'];
		}
		return $zone_id;
	}

	protected function _productIsPhysical($products_id, $attributes) {
		$gv_query = xtc_db_query("select products_model from ".TABLE_PRODUCTS." where products_id = '".xtc_db_input($products_id)."'");
		$gv_result = xtc_db_fetch_array($gv_query);
		if (preg_match('/^GIFT/', $gv_result['products_model'])) {
			return false;
		}

		// check if product is a download
		if(!empty($attributes)) {
			foreach($attributes as $options_id => $options_values_id) {
				$virtual_check_query = xtc_db_query("select count(*) as total from ".TABLE_PRODUCTS_ATTRIBUTES." pa, ".TABLE_PRODUCTS_ATTRIBUTES_DOWNLOAD." pad where pa.products_id = ".(int)$products_id." and pa.options_values_id = ".(int)$options_values_id." and pa.products_attributes_id = pad.products_attributes_id");
				$virtual_check = xtc_db_fetch_array($virtual_check_query);
				$is_download = $virtual_check['total'] > 0;
				if($is_download) {
					return false;
				}
			}
		}
		return true;
	}

	public function stateOrProvinceRequired($iso2) {
		$required_countries = explode(',', self::STATEORPROVINCEREQUIRED);
		$required = in_array($iso2, $required_countries);
		return $required;
	}

	/* =========================================================================================================================== */
	/* =========================================================================================================================== */
	/* == QuickButton helper functions */

	public function doQuickButtonRedirect() {
		if(isset($_SESSION['paypal']['quickbutton'])) {
			if(!headers_sent() && $this->_buttonProductAddedToCart()) {
				unset($_SESSION['paypal']['quickbutton']);
				xtc_redirect(GM_HTTP_SERVER.DIR_WS_CATALOG.'checkout_paypal_prepare.php');
			}
		}
	}

	public function isQuickButtonRedirectRequired($clear_data = false) {
		if(isset($_SESSION['paypal']['quickbutton'])) {
			if($clear_data) {
				unset($_SESSION['paypal']['quickbutton']);
			}
			if($this->_buttonProductAddedToCart()) {
				return GM_HTTP_SERVER.DIR_WS_CATALOG.'checkout_paypal_prepare.php';
			}
		}
		return false;
	}

	protected function _buttonProductAddedToCart() {
		// [new_products_id_in_cart] => 1{2}6{1}3
		// [new_products_id_in_cart] => 2x3
		if(empty($_SESSION['new_products_id_in_cart'])) {
			return false;
		}
		$this->_log("session new product in cart: ".$_SESSION['new_products_id_in_cart']);
		$this->_log("quickbutton_data: ".print_r($_SESSION['paypal']['quickbutton_data'], true));
		$fqpid_cart = new FullyQualifiedProductID((string)$_SESSION['new_products_id_in_cart']);
		$sorted_options = $fqpid_cart->options;
		if(!empty($sorted_options)) {
			ksort($sorted_options);
			$fqpid_cart->options = $sorted_options;
		}
		$button_products_id = (int)$_SESSION['paypal']['quickbutton_data']['products_id'];
		if(!empty($_SESSION['paypal']['quickbutton_data']['id']) && !empty($_SESSION['paypal']['quickbutton_data']['properties_values_ids'])) {
			$type = FullyQualifiedProductID::TYPE_OPTIONSPROPERTIES;
			$options = $_SESSION['paypal']['quickbutton_data']['id'];
			$pcid = $this->_findPropertiesCombisID($_SESSION['paypal']['quickbutton_data']['properties_values_ids']);
		}
		else if(!empty($_SESSION['paypal']['quickbutton_data']['id'])) {
			$type = FullyQualifiedProductID::TYPE_OPTIONS;
			$options = $_SESSION['paypal']['quickbutton_data']['id'];
			$pcid = null;
		}
		else if(!empty($_SESSION['paypal']['quickbutton_data']['properties_values_ids'])) {
			$type = FullyQualifiedProductID::TYPE_PROPERTIES;
			$options = null;
			$pcid = $this->_findPropertiesCombisID($_SESSION['paypal']['quickbutton_data']['properties_values_ids']);
		}
		else {
			$type = FullyQualifiedProductID::TYPE_SIMPLE;
			$options = null;
			$pcid = null;
		}
		if(!empty($options)) {
			ksort($options);
		}
		$fqpid_button = new FullyQualifiedProductID($type, $button_products_id, $options, $pcid);
		$cart_fqpid_string = $fqpid_cart->toInternalRepresentation();
		$button_fqpid_string = $fqpid_button->toInternalRepresentation();
		$this->_log("new in cart: $cart_fqpid_string");
		$this->_log("quickbutton: $button_fqpid_string");
		$products_are_equal = (string)$cart_fqpid_string == (string)$button_fqpid_string;
		$this->_log("new in cart equals quickbutton: ".($products_are_equal ? 'yes' : 'no'));
		return $products_are_equal;
	}

	protected function _findPropertiesCombisID($properties_values_ids) {
		$first_id = array_pop($properties_values_ids);
		$query = "SELECT ppcv1.products_properties_combis_id FROM `products_properties_combis_values` ppcv1 ";
		$join_no = 2;
		foreach($properties_values_ids as $pvid) {
			$query .= "JOIN products_properties_combis_values ppcv".$join_no." ON ppcv".$join_no.".products_properties_combis_id = ppcv1.products_properties_combis_id AND ppcv".$join_no.".properties_values_id = ".(int)$pvid;
			$join_no++;
		}
		$query .= " WHERE ppcv1.properties_values_id = ".(int)$first_id;
		$ppcid = false;
		$result = xtc_db_query($query);
		while($row = xtc_db_fetch_array($result)) {
			$ppcid = $row['products_properties_combis_id'];
		}
		return $ppcid;
	}


}


class GMPayPalException extends Exception {}
class GMPayPalNoTokenException extends GMPayPalException {}
class GMPayPalNoDataException extends GMPayPalException {}
class GMPayPalNoCheckoutDetailsException extends GMPayPalException {}
class GMPayPalTokenExpiredException extends GMPayPalException {}


